// Copyright (C) 2020  Nexedi SA and Contributors.
//                     Kirill Smelkov <kirr@nexedi.com>
//
// This program is free software: you can Use, Study, Modify and Redistribute
// it under the terms of the GNU General Public License version 3, or (at your
// option) any later version, as published by the Free Software Foundation.
//
// You can also Link and Combine this program with other software covered by
// the terms of any of the Free Software licenses or any of the Open Source
// Initiative approved licenses and Convey the resulting work. Corresponding
// source of such a combination shall include the source code for all other
// software used.
//
// This program is distributed WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// See COPYING file for full licensing terms.
// See https://www.nexedi.com/licensing for rationale and options.

package main

import (
	"bufio"
	"context"
	"fmt"
	"io"
	"io/ioutil"
	"os"
	"os/exec"
	"regexp"
	"strings"
	"testing"

	"lab.nexedi.com/kirr/go123/exc"
	"lab.nexedi.com/kirr/go123/xerr"
	"lab.nexedi.com/kirr/neo/go/zodb"
	"lab.nexedi.com/kirr/neo/go/zodb/btree"
)

// TreeGen represents connection to running `treegen trees` server.
//
// Create it with StartTreeGen(zurl).
// - Commit(treeTopology) -> tid
type TreeGen struct {
	zurl	  string
	pysrv     *exec.Cmd      // spawned `treegen trees`
	pyin      io.WriteCloser // input to pysrv
	pyoutRaw  io.ReadCloser  // output from pysrv
	pyout     *bufio.Reader  // buffered ^^^
	treeRoot  zodb.Oid       // oid of the tree treegen works on
	head      zodb.Tid       // last made commit
}

// StartTreeGen spawns `treegen trees` server.
func StartTreeGen(zurl string) (_ *TreeGen, err error) {
	defer xerr.Contextf(&err, "treegen %s: start", zurl)

	// spawn `treegen trees`
	tg := &TreeGen{zurl: zurl}
	tg.pysrv = exec.Command("./testprog/treegen.py", "trees", zurl)
	tg.pyin, err = tg.pysrv.StdinPipe()
	if err != nil {
		return nil, err
	}
	tg.pyoutRaw, err = tg.pysrv.StdoutPipe()
	if err != nil {
		return nil, err
	}
	tg.pyout = bufio.NewReader(tg.pyoutRaw)
	tg.pysrv.Stderr = os.Stderr // no redirection

	err = tg.pysrv.Start()
	if err != nil {
		return nil, err
	}

	// wait for start message; initialize .head and .treeRoot from it
	defer func() {
		if err != nil {
			tg.Close() // ignore error
		}
	}()
	defer xerr.Context(&err, "handshake")
	start, err := tg.pyout.ReadString('\n')
	if err != nil {
		if err == io.EOF {
			err = io.ErrUnexpectedEOF
		}
		return nil, err
	}
	start = strings.TrimSuffix(start, "\n")
	defer xerr.Contextf(&err, "invalid start %q", start)

	// start @<at> tree=<root>
	startRe := regexp.MustCompile(`^start @([^ ]+) root=([^ ]+)$`)
	m := startRe.FindStringSubmatch(start)
	if m == nil {
		return nil, fmt.Errorf("unexpected format")
	}
	tg.head, err = zodb.ParseTid(m[1]) // <at>
	if err != nil {
		return nil, fmt.Errorf("tid: %s", err)
	}
	tg.treeRoot, err = zodb.ParseOid(m[2]) // <root>
	if err != nil {
		return nil, fmt.Errorf("root: %s", err)
	}

	//fmt.Printf("XXX started @%s root=%s\n", tg.head, tg.treeRoot)
	return tg, nil
}

// Close shutdowns the generator.
func (tg *TreeGen) Close() (err error) {
	defer xerr.Contextf(&err, "treegen %s: close", tg.zurl)

	err1 := tg.pyin.Close()
	err2 := tg.pyoutRaw.Close()
	err3 := tg.pysrv.Wait()
	return xerr.Merge(err1, err2, err3)
}

// Commit creates new commit with underlying tree changed to tree topology.
func (tg *TreeGen) Commit(tree string) (_ zodb.Tid, err error) {
	defer xerr.Contextf(&err, "treegen %s: commit %s", tg.zurl, tree)
	fmt.Printf("commit %s ...\n", tree)

	_, err = io.WriteString(tg.pyin, tree + "\n")
	if err != nil {
		return zodb.InvalidTid, err
	}

	reply, err := tg.pyout.ReadString('\n')
	if err != nil {
		return zodb.InvalidTid, err
	}
	reply = strings.TrimSuffix(reply, "\n")

	tid, err := zodb.ParseTid(reply)
	if err != nil {
		return zodb.InvalidTid, fmt.Errorf("invalid reply: %s", err)
	}
	return tid, nil
}


// XGetTree loads LOBtree from zurl@at->obj<root>.
// Tree values must be ZBlk whose data is returned instead of references to ZBlk.
func XGetTree(zurl string, at zodb.Tid, root zodb.Oid) map[int64]string {
	//defer xerr.Contextf(&err, "%s: @%s: get tree %s", zurl, at, root)
	defer exc.Contextf("%s: @%s: get tree %s", zurl, at, root)

	X := exc.Raiseif
	ctx := context.Background()

	// XXX pass IStorage to here instead of zurl?
	zstor, err := zodb.Open(ctx, zurl, &zodb.OpenOptions{ReadOnly: true}); X(err)
	defer zstor.Close() // XXX err

	db := zodb.NewDB(zstor)
	defer db.Close() // XXX err

	// XXX +transaction
	// XXX +defer txn.Abort()

	zconn, err := db.Open(ctx, &zodb.ConnOptions{At: at}); X(err)
	xztree, err := zconn.Get(ctx, root); X(err)
	ztree, ok := xztree.(*btree.LOBTree)
	if !ok {
		exc.Raisef("expected LOBTree, got %s", zodb.ClassOf(xztree))
	}

	err = ztree.PActivate(ctx); X(err)
	defer ztree.PDeactivate()
	zbucket := ztree.FirstBucket()

	kv := make(map[int64]string)
	for zbucket != nil {
		err = zbucket.PActivate(ctx); X(err)
		defer zbucket.PDeactivate()

		for _, __ := range zbucket.Entryv() {
			k  := __.Key()
			xv := __.Value()
			zv, ok := xv.(zBlk)
			if !ok {
				exc.Raisef("[%d] -> %s;  want ZBlk", k, typeOf(xv))
			}

			data, _, err := zv.loadBlkData(ctx); X(err)
			kv[k] = string(data)
		}

		zbucket = zbucket.Next()
	}

	return kv
}

func TestΔBTail(t *testing.T) {
	X := exc.Raiseif

	// XXX test known cases tree1 -> tree2
	testv := []struct { tree1, tree2 string } {
		{"T/B:", "T/B:"},
		{"T/B:", "T/B1:a"},
		// XXX
	}

	work, err := ioutil.TempDir("", "δBTail"); X(err)
	defer func() {
		err := os.RemoveAll(work); X(err)
	}()
	tg, err := StartTreeGen(work + "/1.fs"); X(err)
	defer func() {
		err := tg.Close(); X(err)
	}()

	XXX := func(tree1, tree2 string) {
		// XXX commit tree1
		// XXX load tree1 -> get keys for kv1
		//
		// XXX for all set of {keys} from kv1:
		//	// prepare initial state with {keys} tracked
		//	zconn  = db.open(at)
		//	δbtail = NewΔBTail(zconn.At())
		//	for k in keys:
		//		v, path = tree.VGet(k)
		//		δbtail.Track(path)
		//
		//	commit(->tree2)
		//	δZ = ...
		//
		//	δB = δbtail.Update(δZ)
		//
		//	verify δB has all keys as expected for tracked set

		tid, err := tg.Commit(tree1); X(err)
		tid, err  = tg.Commit(tree2); X(err)

		_ = tid
	}

	for _, test := range testv {
		XXX(test.tree1, test.tree2)
		XXX(test.tree2, test.tree1)
	}
}

func TestΔBTreeAllStructs(t *testing.T) {
	// XXX given (kv1, kv2) test on automatically generated (tree1 -> tree2)
}
